"""Define Transformer for Clustering."""
import math

import torch
import torch.nn as nn
import torch.nn.functional as F

import dgl.geometry as dgl_geo
import segment.utils as seg_utils


class TransformerClustering(nn.Module):

    def __init__(self,
                 num_clusters=4,
                 d_model=512,
                 dropout=0.1,
                 activation="relu",
                 normalize_before=False,
                 detach_src_for_logit=True,
                 l2_normalize_for_fps=True,
                 return_intermediate_dec=False,
                 meanshift_iterations=2,
                 meanshift_step=0.1):
        """Initializes a Transformer for Clustering.

        Args:
          num_clusters: A scalar indicates the number of centroids.
          d_model: A scalar indicates the input channels to Transformer.
          dropout: A `float` indicates the dropout rate.
          activation: A string indicates the type of non-linear activation.
          normalize_before: A `bool` indicates if applying normalization first.
          return_intermediate_dec: A `bool` indicates if return intermediate
            results from decoders.
        """
        super().__init__()
        self.fc = nn.Sequential(
            nn.LayerNorm(d_model),
            nn.ReLU(inplace=True),
            nn.Linear(d_model, d_model, bias=True))

        self._num_clusters = num_clusters
        self._detach_src_for_logit = detach_src_for_logit
        self._l2_normalize_for_fps = l2_normalize_for_fps
        self._meanshift_iterations = meanshift_iterations
        self._meanshift_step = meanshift_step

    def _fill_with_mean(self, src, mask):
        """A helper function to fill invalid entries with mean values.
        """
        bs, sl, cs = src.shape
        if mask is not None:
            mean_src = valid_mean(src, ~mask).unsqueeze(1).type_as(src)
            # Fill padded entries with mean values.
            fill_mask = mask.unsqueeze(2).expand(-1, -1, cs)
            filled_src = torch.where(fill_mask, mean_src.expand(-1, sl, -1), src)
        else:
            mean_src = torch.mean(src, dim=1, keepdim=True).type_as(src)
            filled_src = src

        return filled_src, mean_src

    def _collect_centroids(self, src, kmean_labels, mask, num_clusters):
        """Helper function to collect centroids given kmeans labels.
        """
        bs, sl, cs = src.shape

        if mask is None:
            mask = torch.zeros((bs, sl), dtype=torch.bool, device=src.device)

        # Mask out invalid datas given mask.
        kmean_labels = kmean_labels.masked_fill(mask, num_clusters)

        centroids = seg_utils.segment_mean_nd(src, kmean_labels, 1, 1)

        centroids = centroids[:, :num_clusters, :]

        return centroids

    def _unfold_centroids(self, centroids, kmean_labels, mask):
        """Helper function to collect centroids given kmeans labels.
        """
        bs, ns, cs = centroids.shape

        # Mask out invalid datas given mask.
        if mask is not None:
            kmean_labels = kmean_labels.masked_fill(mask, ns)
        kmean_labels = torch.clamp(kmean_labels, max=ns)

        # Pad centroids with zeros.
        centroids = F.pad(centroids, (0, 0, 0, 1, 0, 0))

        unfold_kmean_labels = kmean_labels.unsqueeze(2).expand(-1, -1, cs)
        unfold_centroids = torch.gather(centroids, 1, unfold_kmean_labels)

        return unfold_centroids

    def _mean_shift(self, src, kmean_labels, mask, num_clusters, iterations, step):
        """Helper function to perform meanshift.
        """
        bs, sl, cs = src.shape
        for _ in range(iterations):
            centroids = self._collect_centroids(
                src, kmean_labels, mask, num_clusters)
            unfold_centroids = self._unfold_centroids(
                centroids, kmean_labels, mask)

            if mask is None:
                mask = torch.zeros((bs, sl), dtype=torch.bool, device=src.device)

            src = torch.where(
                mask.unsqueeze(2).expand(-1, -1, cs),
                src,
                (1 - step) * src + step * unfold_centroids)

        return src

    def _kmeans(self, src, sampled_inds, mask, iterations, metric='l2'):
        """Get centroid by KMeans.
        """
        bs, sl, cs = src.shape
        num_samples = sampled_inds.shape[1]

        src = src.to(torch.float64)

        # Initialize seeds with sampled inds.
        kmean_labels = torch.zeros((bs, sl), dtype=torch.long, device=src.device)
        kmean_labels += num_samples

        sampled_values = torch.arange(num_samples, dtype=torch.long,
                                      device=kmean_labels.device)
        sampled_values = sampled_values.view(1, -1).expand(bs, -1)
        kmean_labels = kmean_labels.scatter(1, sampled_inds, sampled_values)

        # Collect centroids.
        centroids = self._collect_centroids(src, kmean_labels, mask, num_samples)

        for it in range(iterations):

            # Compute distances between datas and centroids.
            if metric == 'l2':
                # Calculate L2 distance.
                sqr_src = torch.sum(src * src, dim=-1).unsqueeze(2)
                sqr_centroids = torch.sum(centroids * centroids, dim=-1).unsqueeze(1)
                src_centroids = torch.einsum('bij,bjk->bik', src, centroids.transpose(1, 2))
                dists = sqr_src + sqr_centroids - 2 * src_centroids
            else:
                # Calculate cosine-similarity distance.
                normed_src = F.normalize(src, dim=-1)
                normed_centroids = F.normalize(centroids, dim=-1)
                dists = 1 - torch.einsum('bij,bjk->bik', normed_src, normed_centroids.transpose(1, 2))

            # Find nearest neighbors.
            kmean_labels = torch.argmin(dists, dim=-1)

            # Update centroids.
            centroids = self._collect_centroids(src, kmean_labels, mask, num_samples)

        return kmean_labels


    def forward(self, src, mask, pos_embed):
        """Feedforward for clustering with Transformer.

        Args:
          src: A `tensor` of shape `[batch_size, source_sequence_length, channels]`.
          mask: A bool `tensor` of shape `[batch_size, sequence_length]`.
          pos_embed: A `tensor` of shape
            `[batch_size, source_sequence_length, channels]`.

        Returns:
          centroids: A `tensor` of shape `[batch_size, num_clusters, channels]`.
          logits: A `tensor` of shape
            `[batch_size, source_sequence_length, num_clusters]`.
          sampled_inds: A `tensor` of shape
            `[batch_size, num_clusters]`.
        """
        bs, sl, cs = src.shape

        # Sample query by Farthest Point Sampling.
        # `centroids` is of shape `[batch_size, target_sequence_length, channels]`.
        filled_src, mean_src = self._fill_with_mean(src, mask)
        padded_src = torch.cat([mean_src, filled_src], dim=1)

        if self._l2_normalize_for_fps:
            sampling_src = F.normalize(padded_src, dim=-1)
        else:
            sampling_src = padded_src

        # NOTE: if the number of valid entries is smaller than
        # num_clusters, the 1st padded entry will be re-sampled.
        sampled_inds = dgl_geo.farthest_point_sampler(
            sampling_src.to(torch.float64),
            self._num_clusters + 1,
            0).long()
        sampled_inds = sampled_inds[:, 1:] - 1
        assert((sampled_inds  >= 0).all())
        unfold_sampled_inds = sampled_inds.unsqueeze(2).expand(-1, -1, cs)
        #sampled_inds = sampled_inds.unsqueeze(2).expand(-1, -1, cs)
        #sampled_src = torch.gather(padded_src, 1, sampled_inds)

        #kmean_labels = self._kmeans(src, sampled_inds[:, :, 0] - 1, mask,
        kmean_labels = self._kmeans(src, sampled_inds, mask,
                                    1, 'l2')

        meanshift_src = self._mean_shift(src, kmean_labels, mask,
                                         self._num_clusters,
                                         self._meanshift_iterations,
                                         self._meanshift_step)
        centroids = torch.gather(meanshift_src, 1, unfold_sampled_inds)

        # Optimize clustering regularization after a fc layer.
        node_features = src.detach() if self._detach_src_for_logit else src
        node_features = self.fc(node_features)

        meanshift_node_features = self._mean_shift(
            node_features, kmean_labels, mask,
            self._num_clusters,
            self._meanshift_iterations,
            self._meanshift_step)
        centroid_features = torch.gather(meanshift_node_features, 1, unfold_sampled_inds)

        # Pick centroids with maximum activations w.r.t `node_features`, where
        # `logits` is of shape
        # `[batch_size, source_sequence_length, target_sequence_length].
        normed_centroid_features = F.normalize(centroid_features, dim=-1)
        normed_node_features = F.normalize(node_features, dim=-1)
        logits = torch.einsum(
            'bij,bjk->bik', normed_node_features, normed_centroid_features.transpose(1, 2))
        logits = logits * 5

        #return centroids, logits, sampled_inds[:, :, 0] - 1
        return centroids, logits, sampled_inds


def valid_mean(x, mask):
     """Compute mean of x given valid mask.

     Args:
         x: A `float` tensor of shape `[batch_size, num_nodes, channels]`.
         mask: A `bool` tensor of shape `[batch_size, num_nodes]`, where
             `True` indicates the entry is valid.

     Returns:
         mean_x: A `float` tensor of shape `[batch_size, channels]`.
     """
     mask = mask.type_as(x).unsqueeze(2)
     sum_mask = torch.clamp(torch.sum(mask, dim=1), min=1)
     masked_x = x * mask
     mean_x = torch.sum(masked_x, dim=1) / sum_mask

     return mean_x

